/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/app.hpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Application class interface.
**/

#ifndef _APP_HPP_
#define _APP_HPP_

#include <set>

#include "base.hpp"
#include "settings.hpp"
#include "interceptor.hpp"
#include "layouter.hpp"
#include "tapper.hpp"

struct argp_state;  // instead of #include <argp.h>.

namespace tapper {

/**
    Application class.

    Application parses command line options and arguments, loads and saves settings, basing on
    configuration, selects and creates interceptor and layouter, creates and runs tapper or prints
    list of known keys or list of known layouts.
**/
class app_t : public object_t {

    public :

        /** Application exit status. **/
        enum exit_status_t {
            ok,                 ///< Everything is ok.
            not_ok,             ///< False, but not an error. Not actually used.
            cmdline_error,      ///< Usage error.
            runtime_error,      ///< Runtime error.
        };

        /** Action to perform. **/
        enum class action_t {
            dflt,               ///< Default action.
            run = dflt,         ///< Run tapper and let it work until SIGINT or SIGTERM, then exit.
            save_settings,      ///< Save settings to GSettings and exit.
            list_keys,          ///< List known keys and exit.
            list_layouts,       ///< List known layouts and exit.
        };

        /** Application exception class. **/
        class error_t : public std::runtime_error {
            public :
                using std::runtime_error::runtime_error;
        }; // class error_t

        typedef std::set< key_t >   key_set_t;
        typedef ptr_t< key_set_t >  key_set_p;

    public :

        app_t( int argc, char * argv[] );
        ~app_t();

    private :

        typedef ptr_t< strings_t >  strings_p;

        // Command line parsing routines:
        static ::error_t parse_opt_or_arg( int key, char * _arg, argp_state * state );
        void   parse_argv( int argc, char * argv[] );
        void   parse_arg( string_t const & arg );
        key_t  parse_key( string_t const & str );
        uint_t parse_layout( string_t const & layout );
        uint_t parse_int( string_t const & str );

        // Low-level actions, called from parsing routines:
        void set_interceptor( settings_t::interceptor_t interceptor );
        void set_layouter( settings_t::layouter_t layouter );
        void set_bell( settings_t::bell_t bell );
        void set_action( action_t action );
        void load_settings();
        void save_settings();

        // Top-level actions:
        void list_keys();
        void list_layouts();
        void run();

        void print_intro();

        string_t            layout_name( uint_t code );

        interceptor_t &     interceptor();
        layouter_t &        layouter();
        settings_t::bell_t  bell();

    private :

        char const * const  prog_name;                      ///< Program name.

        settings_t          _settings;                      ///< Current settings.
        bool                _load_settings { ENABLE_GNOME };
            ///< If true, the app will load setting from GSettings, otherwise it won't.
        bool                _show_taps { false };
            ///< If true, the app will print each detected tap, otherwise it won't
        bool                _quiet { false };
            ///< If true, the app won't print intro message.
        bool                _syslog { false };
            ///< If true, copy error and warning messages to syslog.
        action_t            _action { action_t::dflt };     ///< Action to perform.
        key_set_p           _used_keys;
            ///< Set of keys used in the command line to detect keys assigned more than once.

        interceptor_p       _interceptor;
        layouter_p          _layouter;

}; // class app_t

}; // namespace tapper

#endif // _APP_HPP_

// end of file //
