/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/base.cpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Tapper basic infrastructure implementation.
**/

#include "base.hpp"

#include <iomanip>
#include <iostream>
#include <map>
#include <mutex>

#include "posix.hpp"

namespace tapper {

#if defined( DEBUG )

static thread_local uint_t _level = 0;

static
uint_t
gen_no(
    char const * type
) {
    // We do not have much contention in creating objects, so it is ok to use mutex.
    static std::mutex mutex;
    static std::map< char const *, uint_t > counters;
    std::lock_guard< std::mutex > lock( mutex );
    return ++ counters[ type ];
};

#endif

// -------------------------------------------------------------------------------------------------
// object_t
// -------------------------------------------------------------------------------------------------

#if defined( DEBUG )

object_t::object_t(
    char const * type
) :
    _no( gen_no( type ) ),
    _id( STR( chop( type, "_t" ) << "#" << _no ) )
{
}; // ctor

object_t const *
object_t::_this_(
) const {
    return this;
}; // _this_

uint_t
object_t::_no_(
) const {
    return _no;
}; // _no_

string_t
object_t::_id_(
) const {
    return _id;
}; // _id_

object_t const *
_this_(
) {
    return nullptr;
}; // _this_

#endif

// -------------------------------------------------------------------------------------------------
// xprint functions
// -------------------------------------------------------------------------------------------------

#if defined( DEBUG )

    void
    dprint(
        object_t const *    obj,
        char const *        file,   ///< Source file name.
        int                 line,   ///< Source file line number.
        string_t const &    mesg    ///< Message to print.
    ) {
        /*
            Note: We construct a full message first by calling `STR`, and then print it to `cout` by
            calling operator `<<`. This is intentional. Otherwise, if `STR` is omitted, every piece
            is printed separetely. It causes output from parallel threads interleaved.
        */
        auto no = posix::thread_t::no();
        eprint( priority_t::debug, STR( ""
            << "[" << std::setw( 1 ) << ( no == 0 ? "." : STR( no ) ) << "] "
            << std::setw( 16 )
            << ( file ? STR( posix::base_name( file ) + ":" << std::setw( 3 ) << line ) : "" )
            << ":"
            << string_t( _level, ' ' )
            << ( obj ? obj->_id_() + ": " : "" )
            << mesg
        ) );
    }; // dprint

#endif

static auto priority { priority_t( 0 ) };

void eprint( priority_t level, string_t const & message ) {
    char prefix = '?';
    switch ( level ) {
        case priority_t::debug : {
            prefix = '#';
        } break;
        case priority_t::info : {
            prefix = 'I';
        } break;
        case priority_t::warning : {
            prefix = 'W';
        } break;
        case priority_t::error : {
            prefix = 'E';
        } break;
    };
    if ( level <= priority ) {
        ::syslog( int( level ), "%c %s", prefix, message.c_str() );
    };
    std::cerr << STR( program_invocation_short_name << ": " << prefix << " " << message << "\n" );
}; // eprint

void oprint( string_t const & message ) {
    std::cout << message << "\n";
}; // oprint

void syslog_min_priority( priority_t level ) {
    priority = level;
};

// -------------------------------------------------------------------------------------------------
// tracer_t
// -------------------------------------------------------------------------------------------------

#if defined( DEBUG )

_func_tracer_t::_func_tracer_t(
    object_t const *    obj,
    char const *        file,
    int                 line,
    char const *        func
) :
    _obj( obj ),
    _file( file ),
    _line( line ),
    _func( func )
{
    dprint( _obj, _file, _line, STR( func << " {" ) );
    ++ _level;
}; // ctor

_func_tracer_t::~_func_tracer_t(
) {
    -- _level;
    dprint( nullptr, _file, _line, "}" );
    //    ^^^^^^^ Do not prefix "}" with object id.
}; // dtor

#endif

}; // namespace tapper

// end of file //
