/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/base.hpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Basic infrastructure interface.

    Basic infrastructure provides facilities for error handling, logging, and debugging.
**/

#ifndef _BASE_HPP_
#define _BASE_HPP_

#include "config.h"

#include <exception>

#include <syslog.h>

#include "types.hpp"

/// Top-level namespace.
namespace tapper {

/**
    Base class for all other classes.

    Objects are not copyable.

    It provides debugging facilities: each object has type (object type is used for constructing
    object id but is not saved as data member), number, and id.

    Object type is a string, it should be specified in the constructor. It is expected `OBJECT_T()`
    is used for it:

    @code {.cpp}
        class class_t : public object_t {
            public :
                class_t() : OBJECT_T() {
                    ...
                };
                ...
        };
    @endcode

    Object number is assigned automatically, sequentially within the type. The first created object
    of the type receives no 1, the next created object of the same type receives number 2, and so
    on. Numbers of deleted objects are not reused.

    Object id is constructed from object type and object number, and looks like "object#1".

    All these stuff is used for debugging purposes solely.

**/
class object_t {

    // Objects are non-copyable.
    object_t( object_t const & ) = delete;
    object_t & operator =( object_t const & ) = delete;

    #if defined( DEBUG )

        public :

            explicit object_t( char const * type );
            virtual ~object_t() = default;

            object_t const * _this_() const;
            uint_t           _no_()   const;
            string_t         _id_()   const;

        private :

            uint_t const    _no;    ///< Object number, unique within the type.
            string_t const  _id;    ///< Object id, unique within the program.

    #else

        public :

            explicit object_t( char const * type ) {};
            virtual ~object_t() = default;

    #endif

}; // object_t

/**
    Calls `object_t` constructor.

    Usage:

    @code
        class something_t : public object_t {
            public :
                something_t::something_t() : OBJECT_T() {
                    ...
                };
                ...
        };
    @endcode
**/
#if defined( DEBUG )
    #define OBJECT_T()  object_t( __FUNCTION__ )
#else
    #define OBJECT_T()  object_t( nullptr )
#endif

/**
    `_this_` global function.

    The idea is that `_this_()` is accessible everywhere. Inside a class it is a `objec_t::_this_`
    (which returns `this`), outside of a class `_this_` is global function (which return
    `nullptr`), so `DBG` can detect if it is called from a class or not.
**/
#if defined( DEBUG )
    object_t const * _this_();
#endif

/**
    Provides `_this_` definition for static methods.

    `DBG` rely on `_this_()`, which could be either `object_t` method or global function. However,
    it does not work in static methods since `_this_` method is defined but not accessible. `THIS`
    macro provides `_this_` definition for static methods:

    @code
    void a_static_method() {
        THIS( pointer-to-this-object );
        ...
        ERR( "message" );
    };
    @endcode
**/
#define THIS( OBJ ) auto _this_ = [ OBJ ] () { return OBJ; }; static_cast< void >( _this_ );

/**
    Message priority level.

    These priority levels are subset of syslog priority levels.

    @note The higher priority, the less priority numerical value. E. g. `error < debug`.
**/
enum class priority_t {
    error   = LOG_ERR,
    warning = LOG_WARNING,
    info    = LOG_INFO,
    debug   = LOG_DEBUG,
}; // enum priority_t

#if defined( DEBUG )
    /// `DBG` guts. Do not use it directly.
    void dprint( object_t const * obj, char const * file, int line, string_t const & mesg );
#endif

/// `INF`, `WRN`, and `ERR` guts. Do not use it directly.
void eprint( priority_t level, string_t const & message );

/// `OUT` guts. Do not use it directly.
void oprint( string_t const & message );

/**
    Set minimum priority level for sending messages to syslog.

    All the messages (errors, warnings, informational, debug) are printed to standard error stream
    (debug messages are printed only in debug build, of course). Messages with specified priority
    and higher will be *additionally* send to syslog.
**/
void syslog_min_priority( priority_t level );

/**
    Prints debug message.

    Usage:
    @code
        DBG( "Opening file " << q( path ) << "..." );
    @endcode

    In case of debug build, the result includes the program name, thread number, source location
    (file name and line number), and object (type and number) the message is printed from. For
    example, the statement above prints:

    @verbatim
    tapper: # [.]       li.cpp:199:context#1: Opening file ‘/dev/input/event1’...
    @endverbatim

    where `tapper` is the program name, `#` is the message priority (`#` stands for debug, `W` for
    warning, `E` for error), `[.]` is thread number (dot denotes primary thread, other threads will
    be numbered starting from 1), `li.cpp:199` is the file name and line number, `context#1` is
    type and number of object, the rest is the `DBG` arguments.

    In case of production build the macro prints no message (and produces no code).

    If `DBG` is called from static method, use `THIS` macro before calling `DBG`, otherwise
    the compiler will generate error:

    @verbatim
    file.cpp:156:18: error: cannot call member function ‘const tapper::object_t*
        tapper::object_t::_this_() const’ without object
    @endverbatim

**/
#define DBG( _ARGS_ ) \
    DBG_( _this_(), _ARGS_ )

/// `DBG` guts. Do not use it directly, use `DBG` instead.
#if defined( DEBUG )
    #define DBG_( _OBJ_, _ARGS_ ) dprint( _OBJ_, __FILE__, __LINE__, STR( _ARGS_ ) )
#else
    #define DBG_( _OBJ_, _ARGS_ ) void( 0 )
#endif

/**
    Prints informational message.
**/
#define INF( _ARGS_ ) \
    INF_( _this_(), _ARGS_ )

/// `INF` guts. Do not use it directly, use `INF` instead.
#define INF_( _OBJ_, _ARGS_ ) {                                                                    \
    string_t _MSG_ = STR( _ARGS_ );                                                                \
    eprint( priority_t::info, _MSG_ );                                                             \
}

/**
    Prints warning message.
**/
#define WRN( _ARGS_ ) \
    WRN_( _this_(), _ARGS_ )

/// `WRN` guts. Do not use it directly, use `WRN` instead.
#define WRN_( _OBJ_, _ARGS_ ) {                                                                    \
    string_t _MSG_ = STR( _ARGS_ );                                                                \
    eprint( priority_t::warning, _MSG_ );                                                          \
}

/**
    Constructs a message from `_ARGS_` argument and throws an exception of `error_t` class. In case
    of debug build, it prints debug message using `DBG` macro just before throwing the exception.

    @code
        ERR( "Out of range " << range << "." );
    @endcode
**/
#define ERR( _ARGS_, ... ) \
    ERR_( _this_(), _ARGS_, ## __VA_ARGS__ )

/// `ERR` guts. Do not use it directly, use `ERR` instead.
#define ERR_( _OBJ_, _ARGS_, ... ) {                                                               \
    string_t _MSG_ = STR( _ARGS_ );                                                                \
    DBG_( _OBJ_, "Throwing: " << _MSG_ );                                                          \
    throw error_t( _MSG_, ## __VA_ARGS__ );                                                        \
}

/**
    TODO
**/
#define OUT( _ARGS_ ) \
    oprint( STR( _ARGS_ ) )

/**
    Catches all exceptions, prints a warning if an exception was caught.

    The macro is used primarily in destructors:

    @code
        something_t::~something_t(
        ) {
            CATCH_ALL( close(); );  // Avoid throwing exceptions from destructor.
        };
    @endcode
**/
#define CATCH_ALL( _STATEMENT_ ) \
    CATCH_ALL_( _this_(), _STATEMENT_ )

/// `CATCH_ALL` guts. Do not use it directly, use `CATCH_ALL` instead.
#define CATCH_ALL_( _OBJ_, _STATEMENT_ ) {                                                         \
    try {                                                                                          \
        _STATEMENT_;                                                                               \
    } catch ( std::exception const & _ex_ ) {                                                      \
        WRN_( _OBJ_, _ex_.what() );                                                                \
    } catch ( ... ) {                                                                              \
        WRN_( _OBJ_, "Unknown exception occurred" );                                               \
    };                                                                                             \
}

#if defined( DEBUG )

    class _func_tracer_t {
        public :
            _func_tracer_t( object_t const * obj, char const * file, int line, char const * func );
            ~_func_tracer_t();
        private :
            object_t const * const  _obj;
            char const * const      _file;
            int const               _line;
            char const * const      _func;
    }; // class _func_tracer_t

#endif

/**
    Traces entering and leaving a scope.

    Usage:

    @code
        void
        function_or_method(
        ) {
            TRACE();
            ...
        };
    @endcode

    The example above in case of debug build will print something like this:

    @verbatim
    tapper: # [.]        file.cpp:165:function_or_method {
    ...
    tapper: # [.]        file.cpp:165:}
    @endverbatim

    Left brace `{` denotes entering a scope, right brace `}` — leaving a scope. All the debug
    messages printed in the scope and nested traces will be indented accordingly to the nesting
    level.

    In prodiction build macro prints noting (and produces no code).
**/
#define TRACE() \
    TRACE_( _this_() )

/// `TRACE` guts. Do not use it directly, use `TRACE` instead.
#if defined( DEBUG )
    #define TRACE_( _OBJ_ ) \
        _func_tracer_t _tracer_ ## __COUNTER__( _OBJ_, __FILE__, __LINE__, __FUNCTION__ )
#else
    #define TRACE_( _OBJ_ )
#endif

}; // namespace tapper

#endif // _BASE_HPP_

// end of file //
