/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/gnome.cpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Gnome wrappers implementation.
**/

#include "gnome.hpp"

#include <glibmm/error.h>
#include <json-glib/json-glib.h>

namespace tapper {
namespace gnome {

// -------------------------------------------------------------------------------------------------
// error_t
// -------------------------------------------------------------------------------------------------

// TODO: Drop it.
class error_t : public std::runtime_error {
    public :
        typedef error_t            myself_t;
        typedef std::runtime_error parent_t;
        explicit error_t( string_t const & message, GError * error = nullptr );
        error_t( myself_t const & other );
        virtual ~error_t();
        myself_t & operator =( myself_t const & other ) = delete;
        bool matches( GQuark domain, gint code ) const;
    private :
        GError * _error;
}; // class error_t

error_t::error_t(
    string_t const & message,
    GError *         error
) :
    parent_t( message ),
    _error( error )
{
}; // ctor

error_t::error_t(
    myself_t const & other
) :
    parent_t( other.what() ),
    _error( g_error_copy( other._error ) )
{
};

error_t::~error_t(
) {
    g_error_free( _error );
}; // dtor

bool
error_t::matches(
    GQuark  domain,
    gint    code
) const {
    return g_error_matches( _error, domain, code );
};

// -------------------------------------------------------------------------------------------------
// Deleters.
// -------------------------------------------------------------------------------------------------

template< typename value_t >
struct deleter_t {
    void operator ()( value_t * ptr ) const {
        g_object_unref( ptr );
    };
};

template<>
struct deleter_t< GVariant > {
    void operator ()( GVariant * ptr ) const {
        g_variant_unref( ptr );
    };
};

template<>
struct deleter_t< GVariantBuilder > {
    void operator ()( GVariantBuilder * ptr ) const {
        g_variant_builder_unref( ptr );
    };
};

// -------------------------------------------------------------------------------------------------
// input_source_manager_t
// -------------------------------------------------------------------------------------------------

static char const * const bus_name = "org.gnome.Shell";
static char const * const obj_path = "/org/gnome/Shell";

input_source_manager_t::input_source_manager_t(
) :
    OBJECT_T()
{
    _reset();
};

input_source_manager_t::~input_source_manager_t(
) {
};

/**
    Resets D-Bus connection to Gnome Shell.
**/
void
input_source_manager_t::_reset(
) {
    try {
        _bus = Gio::DBus::Proxy::create_for_bus_sync(
            Gio::DBus::BUS_TYPE_SESSION,
            bus_name, obj_path, bus_name,
            Glib::RefPtr< Gio::DBus::InterfaceInfo >(),
            Gio::DBus::PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES
                | Gio::DBus::PROXY_FLAGS_DO_NOT_CONNECT_SIGNALS
        );
    } catch ( Glib::Exception const & ex ) {
        ERR( "Can't open bus " << q( bus_name ) << ": " << ex.what() );
    };
};

/**
    Evaluates jscript code in Gnome Shell context. In case of errors, throws an exception.
**/
string_t                        /// @return Result of evaluation.
input_source_manager_t::_eval(
    string_t const & jscript    ///< jscript source code to evaluate.
) {
    static char const * const method = "Eval";
    auto argument = Glib::Variant< std::tuple< string_t > >::create( jscript ); /*
        There is a memory leak in Glib::Variant,
        see <https://gitlab.gnome.org/GNOME/glibmm/issues/48>.
    */
    DBG( "call: " << method << "( " << q( jscript ) << " )" );
    auto reply = _bus->call_sync( method, argument );
    //~ if ( not g_variant_check_format_string( reply.get(), format, TRUE ) ) {
        //~ ERR(
            //~ "Can't parse Gnome Shell response: "
            //~ << "method `" << method << "' returned response which does not match "
            //~ << "format `" << format << "'"
        //~ );
    //~ };
    auto ok =
        Glib::VariantBase::cast_dynamic< Glib::Variant< bool     > >( reply.get_child( 0 ) ).get();
    auto str =
        Glib::VariantBase::cast_dynamic< Glib::Variant< string_t > >( reply.get_child( 1 ) ).get();
    if ( ! ok ) {
        ERR( str );
    };
    DBG( "repl: " << q( str ) );
    return str;
};

/**
    Returns list of input sources, currently configured in Gnome Shell.
**/
strings_t
input_source_manager_t::keyboard_layouts(
) {
    typedef std::unique_ptr< JsonParser, deleter_t< JsonParser > > json_parser_p;
    strings_t layouts;
    // TODO: Join all names into a string which could be parsed without Json?
    string_t  reply = _eval(
        "let layouts = [];"
        "const input_sources = imports.ui.status.keyboard.getInputSourceManager().inputSources;"
        "for ( let i in input_sources ) {"
            "const source = input_sources[ i ];"
            "layouts[ source.index ] = source.displayName;"
        "};"
        "layouts;"
    );
    json_parser_p parser( json_parser_new() );
    GError * error = nullptr;
    json_parser_load_from_data( parser.get(), DATA( reply ), & error );
    if ( error ) {
        ERR( error->message << ": " << q( reply ), error );
    };
    static char const * const message =
        "Can't parse Gnome Shell response: array of strings is expected but received";
    auto root = json_parser_get_root( parser.get() );
    if ( not JSON_NODE_HOLDS_ARRAY( root ) ) {
        ERR( message << " " << q( reply ) << "." );
    };
    auto array = json_node_get_array( root );
    auto len = json_array_get_length( array );
    for ( size_t i = 0; i < len; ++ i ) {
        auto elem = json_array_get_element( array, i );
        if ( not JSON_NODE_HOLDS_VALUE( elem ) ) {
            ERR( message << " " << q( reply ) << "." );
        };
        auto name = json_node_get_string( elem );
        if ( not name ) {
            ERR( message << " " << q( reply ) << "." );
        };
        layouts.push_back( name );
    };
    return layouts;
};

/**
    Activates a layout.
**/
void
input_source_manager_t::activate_layout(
    uint_t  idx,    ///< Index of layout to activate.
    bool    bell    ///< If true, ring a bell.
) {
    /*
        If Gnome Shell is restarted, activating layout fails with error
        GDBus.Error:org.freedesktop.DBus.Error.ServiceUnknown. Resetting connection helps to
        restore normal functioning.

        If resetting connection din't help — just ignore the error. If Gnome Shell crashes,
        restarting may require a long time. Keyboard layout switching will not work for some time,
        but when Gnome Shell is up, Tapper will work again.
    */
    size_t attempts = 0;
    start:
    try {
        static char const * const play_bell =
            "const player = global.display.get_sound_player();"
            "player.play_from_theme( 'bell', \"Keyboard layout selected\", null );";
        // TODO: eval async?
        auto reply = _eval(
            STR(
                "const idx = " << idx << ";"
                "const sources = imports.ui.status.keyboard.getInputSourceManager().inputSources;"
                "if ( idx in sources ) {"
                    "sources[ idx ].activate( false );"
                    << ( bell ? play_bell : "" ) <<
                    "true;"
                "} else {"
                    "false;"
                "};"
            )
        );
        if ( reply == "true" ) {
            // Ok.
        } else if ( reply == "false" ) {
            WRN( "There is no layout #" << ( idx + 1 ) << "." );
        } else {
            ERR( reply );
        };
    } catch ( Glib::Error const & ex ) {
        if ( ex.matches( G_DBUS_ERROR, G_DBUS_ERROR_SERVICE_UNKNOWN ) ) {
            if ( not attempts ) {
                WRN( ex.what() );
                WRN( "Looks like Gnome Shell is not active, trying to reset connection…" );
                _reset();
                ++ attempts;
                goto start;
            } else {
                WRN( "Resetting connection didn't help, layout is not activated." );
            };
        };
        ERR( ex.what() );
    };
};

}; // namespace gnome
}; // namespace tapper

// end of file //
