/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/interceptor-x.cpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    `interceptor::x_t` class implementation.
**/

#include "interceptor-x.hpp"

#include <limits>

#include <cassert>

#include <byteswap.h>       // bswap_64
#include <X11/Xproto.h>     // xEvent

namespace tapper {
namespace interceptor {

// -------------------------------------------------------------------------------------------------
// x_t
// -------------------------------------------------------------------------------------------------

x_t::x_t(
) :
    OBJECT_T(),
    _context( std::bind( & x_t::on_intercept, this, std::placeholders::_1 ) )
{
}; // ctor

x_t::~x_t(
) {
}; // dtor

string_t
x_t::type(
) {
    return "XRecord";
}; // type

key_range_t const &
x_t::key_range(
) {
    if ( not _key_range ) {
        // Get Xkb key range:
        auto xkb = kb().desc()->key_range();
        // Convert Xkb key range to libinput key range:
        auto li = key_range_t( x::key( xkb.min ), x::key( xkb.max ) );
        // Exclude libinput's KEY_RESERVED from range:
        if ( li.min == 0 ) {
            li.min = 1;
        };
        /*
            Extend key range with button codes. X button codes are sequential from `x::button_min` to
            `x::button_max`, but corresponding libinput key codes may or may not be sequential, so
            range `key_range_t( x::button( x::button_min ), x::button( x::button_max ) )` may be
            invalid. Let's
        */
        for ( uint_t b = x::button_min; b <= x::button_max; ++ b ) {
            li += x::button( b );
        };
        assert( li.min != 0 );
        _key_range.reset( new key_range_t( li ) );
    };
    // Return libinput key range:
    return * _key_range.get();
}; // key_range

key_names_t const &
x_t::key_names(
) {
    if ( not _key_names ) {
        key_names_t names;
        for ( auto & it : kb().desc()->key_names() ) {
            auto key = x::key( it.second );
            if ( key ) {
                auto inserted = names.insert( { it.first, key } ).second;
                assert( inserted );
            } else {
                WRN( "Xkb reported unexpected key code " << it.second << "." );
            };
        };
        // Add mouse button names:
        for ( auto b = x::button_min; b <= x::button_max; ++ b ) {
            auto name = "BTN" + std::to_string( b );
            auto inserted = names.insert( { name, x::button( b ) } ).second;
            if ( not inserted ) {
                WRN(
                    "Can't use name " << q( name ) << " for mouse button "
                        << "since Xkb reported key with such name."
                );
            };
        };
        _key_names.reset( new key_names_t( names ) );
    };
    return * _key_names.get();
}; // key_names

key_codes_t const &
x_t::key_codes(
) {
    if ( not _key_codes ) {
        key_codes_t codes = reverse( key_names() );
        _key_codes.reset( new key_codes_t( codes ) );
    };
    return * _key_codes.get();
}; // key_codes

key_t
x_t::key(
    string_t const & name
) {
    auto & names = key_names();
    auto it = names.find( uc( name ) );
    if ( it == names.end() ) {
        return 0;
    };
    return it->second;
}; // key

void
x_t::start(
) {
    _context.enable();
}; // start

void
x_t::stop(
) {
    _context.disable();
}; // stop

string_t
x_t::_key_name(
    key_t key
) {
    auto & codes = key_codes();
    auto it = codes.find( key );
    if ( it == codes.end() ) {
        return "";
    };
    return it->second;
}; // _key_name

x::kb_t &
x_t::kb(
) {
    if ( not _kb ) {
        _kb.reset( new x::kb_t( _context.display() ) );
    };
    return * _kb.get();
}; // kb

void
x_t::on_intercept(
    XRecordInterceptData const * data
) {
    TRACE();
    if ( data->category == XRecordFromServer ) {

        /*
            Decoding data described in:
            http://www.x.org/releases/X11R7.7/doc/xproto/x11protocol.html#events:input
        */
        if ( data->data_len * 4 < sizeof( xEvent ) ) {
            // `data_len` is size of data in 4-byte units.
            WRN( "Intercepted data is too short." );
            return;
        }; // if
        auto event = reinterpret_cast< xEvent const * >( data->data );
        auto get_x_event_type = [ data, event ] () {
            auto type = event->u.u.type;
            static_assert( sizeof( type ) == 1 );   // No need to swap bytes.
            return type;
        };
        auto get_x_event_key = [ data, event ] () {
            auto key = event->u.u.detail;
            static_assert( sizeof( key ) == 1 );    // No need to swap bytes.
            return key;
        };
        auto get_x_event_button = get_x_event_key;
        auto get_x_event_time = [ data, event ] () {
            auto time = event->u.keyButtonPointer.time;
            if ( data->client_swapped ) {
                WRN( "Swapping bytes…" );            // TODO: Warn once?
                static_assert( sizeof( time ) == 4 );
                time = bswap_32( time );
            };
            return time;
        };

        auto type = get_x_event_type();
        // My experiments show that data->server_time == event->u.keyButtonPointer.time.
        switch ( type ) {
            case KeyPress      :
            case KeyRelease    : {
                auto key = x::key( get_x_event_key() );
                if ( key ) {
                    _on_event( {
                        .time  = get_x_event_time(),
                        .key   = key,
                        .state = key_state( type == KeyPress ),
                    } );
                };
            } break;
            case ButtonPress   :
            case ButtonRelease : {
                auto key = x::button( get_x_event_button() );
                if ( key ) {
                    _on_event( {
                        .time  = get_x_event_time(),
                        .key   = key,
                        .state = key_state( type == ButtonPress ),
                    } );
                };
            } break;
        }; // switch

    }; // if
}; // on_intercept

}; // namespace interceptor
}; // namespace tapper

// end of file //
