/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/interceptor.hpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    `interceptor_t` abstract class interface.
**/

#ifndef _INTERCEPTOR_HPP_
#define _INTERCEPTOR_HPP_

#include <functional>

#include "base.hpp"

namespace tapper {

// -------------------------------------------------------------------------------------------------
// interceptor_t
// -------------------------------------------------------------------------------------------------

class interceptor_t;
typedef ptr_t< interceptor_t > interceptor_p;

/**
    Interceptor abstract interface.

    Interceptor is a component which listens for keyboard events and calls specified function on
    each event.
**/
class interceptor_t {

    public :

        /** Interceptor exceptions. **/
        class error_t : public std::runtime_error {
            public :
                using std::runtime_error::runtime_error;
        }; // class error_t

        /** Key state: pressed or released. **/
        enum class key_state_t {
            released,
            pressed,
        };

        /**
            Converts boolean value to `key_state_t` value. `true` converted to
            `key_state_t::pressed`, `false` — to `key_state_t::released`.
        */
        key_state_t key_state( bool pressed );

        /** Keyboard event. **/
        struct event_t {
            time_t      time;   ///< Time when the event occurred.
            key_t       key;    ///< Key which state was changed.
            key_state_t state;  ///< New state of the key.
        };

        /**
            Type of function called on every keyboard event.
        **/
        typedef std::function< void( event_t const & ) > on_event_t;

    public :

        virtual ~interceptor_t() = default;

        /** Returns the interceptor type name. **/
        virtual string_t type() = 0;

        virtual key_range_t const & key_range() = 0;
        virtual key_names_t const & key_names() = 0;
        virtual key_codes_t const & key_codes() = 0;

        virtual string_t            key_name( key_t key );
        virtual key_t               key( string_t const & name ) = 0;
            ///< Converts key name to key code.

        virtual void set_on_event( on_event_t on_event );
        virtual void start() = 0;   ///< Starts intercepting.
        virtual void stop()  = 0;   ///< Stops intercepting.

    protected :

        virtual string_t    _key_name( key_t key );

    protected :

        on_event_t  _on_event;      ///< Function to call on every intercepted event.

}; // class interceptor_t

/**
    @namespace tapper::interceptor

    Concrete interceptors defined in this namespace.
**/

}; // namespace tapper

#endif // _INTERCEPTOR_HPP_

// end of file //
