/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/strings.cpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    String manipulating utilities implementation.
**/

#include "string.hpp"

#include <algorithm>
#include <functional>

#include <cassert>

namespace tapper {

#if _NDEBUG
    // No tests in production build.
    #define TEST( _CODE_ )
#else
    namespace {
        struct Test {
            explicit Test( std::function< void() > test ) {
                test();
            };
        };
    };
    #define TEST( _CODE_ )          TEST_( __COUNTER__, _CODE_ )
    #define TEST_( _N_, _CODE_ )    TEST__( _N_, _CODE_ )
    #define TEST__( _N_, _CODE_ )   static Test t##_N_( [](){ _CODE_ } )
#endif // _NDEBUG

/**
    Perl-like `split`.
**/
strings_t
split(
    char              chr,  ///< Delimiter.
    string_t const &  str,  ///< String to split.
    size_t            limit ///< Maximal number of strings to return.
) {
    strings_t res;
    if ( limit > 0 ) {
        size_t size = str.size();
        size_t p = 0;
        while ( limit > 1 and p < size ) {
            // cppcheck-suppress shadowFunction
            auto q = str.find( chr, p );
            if ( q == npos ) {
                q = size;
            };
            res.push_back( str.substr( p, q - p ) );
            p = q + 1;
            limit -= 1;
        }; // while
        if ( p <= size ) {
            res.push_back( str.substr( p, size - p ) );
        }; // if
    }; // if
    return res;
}; // split

TEST(
    assert( split( '=', "abc", 4 ) == strings_t( { "abc" } ) );
    assert( split( '=', "", 4 ) == strings_t( { "" } ) );
    assert( split( '=', "", 0 ) == strings_t( {} ) );

    assert( split( '=', "a=b=c", 0 ) == strings_t( {} ) );
    assert( split( '=', "a=b=c", 1 ) == strings_t( { "a=b=c" } ) );
    assert( split( '=', "a=b=c", 2 ) == strings_t( { "a", "b=c" } ) );
    assert( split( '=', "a=b=c", 3 ) == strings_t( { "a", "b", "c" } ) );
    assert( split( '=', "a=b=c", 4 ) == strings_t( { "a", "b", "c" } ) );

    assert( split( '=', "==", 0 ) == strings_t( {} ) );
    assert( split( '=', "==", 1 ) == strings_t( { "==" } ) );
    assert( split( '=', "==", 2 ) == strings_t( { "", "=" } ) );
    assert( split( '=', "==", 3 ) == strings_t( { "", "", "" } ) );
    assert( split( '=', "==", 4 ) == strings_t( { "", "", "" } ) );

    assert( split( '=', "эюя", 4 ) == strings_t( { "эюя" } ) );

    assert( split( '=', "э=ю=я", 0 ) == strings_t( {} ) );
    assert( split( '=', "э=ю=я", 1 ) == strings_t( { "э=ю=я" } ) );
    assert( split( '=', "э=ю=я", 2 ) == strings_t( { "э", "ю=я" } ) );
    assert( split( '=', "э=ю=я", 3 ) == strings_t( { "э", "ю", "я" } ) );
    assert( split( '=', "э=ю=я", 4 ) == strings_t( { "э", "ю", "я" } ) );
);

/**
    Perl-like `join`.
**/
string_t
join(
    string_t const &  joiner,
    strings_t const & strings
) {
    string_t r;
    for ( size_t i = 0, end = strings.size(); i < end; ++ i ) {
        r += ( i == 0 ? "" : joiner );
        r += strings[ i ];
    };
    return r;
};

TEST(
    assert( join( ", ", strings_t( { "a", "b", "c" } ) ) == "a, b, c" );
    assert( join( ", ", strings_t( { "a", "b" } ) ) == "a, b" );
    assert( join( ", ", strings_t( { "a" } ) ) == "a" );
    assert( join( ", ", strings_t( { } ) ) == "" );

    assert( join( "", strings_t( { "a", "b", "c" } ) ) == "abc" );
    assert( join( "", strings_t( { "a", "b" } ) ) == "ab" );
    assert( join( "", strings_t( { "a" } ) ) == "a" );
    assert( join( "", strings_t( { } ) ) == "" );

    assert( join( ", ", strings_t( { "", "", "" } ) ) == ", , " );
    assert( join( "", strings_t( { "" } ) ) == "" );

    assert( join( ", ", strings_t( { "э", "ю", "я" } ) ) == "э, ю, я" );
    assert( join( ", ", strings_t( { "э", "ю" } ) ) == "э, ю" );
    assert( join( ", ", strings_t( { "э" } ) ) == "э" );
    assert( join( ", ", strings_t( { } ) ) == "" );
);

/**
    Searches a string vector for a string.

    @return Index of the found string, or `npos` if the string is not found.
**/
size_t
find(
    string_t const &  needle,   ///< String to search for.
    strings_t const & haystack  ///< Strings to search in.
) {
    auto it = std::find( haystack.begin(), haystack.end(), needle );
    return it == haystack.end() ? npos : it - haystack.begin();
}; // find

/**
    Extracts specified number of characters from the beginning of a string. If the string is
    shorter than requested number of characters, all characters of the string are returned.

    @return String containing extracted characters.
**/
string_t
left(
    string_t const & str,   ///< String to extract characters from.
    size_t           len    ///< Number of characters to extract.
) {
    return str.substr( 0, len );
}; // left

TEST(
    assert( left( "", 4 ) == "" );

    assert( left( "abc", 0 ) == "" );
    assert( left( "abc", 1 ) == "a" );
    assert( left( "abc", 2 ) == "ab" );
    assert( left( "abc", 3 ) == "abc" );
    assert( left( "abc", 4 ) == "abc" );

    if ( ENABLE_GNOME ) {
        assert( left( "эюя", 0 ) == "" );
        assert( left( "эюя", 1 ) == "э" );
        assert( left( "эюя", 2 ) == "эю" );
        assert( left( "эюя", 3 ) == "эюя" );
        assert( left( "эюя", 4 ) == "эюя" );
    }; // ENABLE_GNOME
);

/**
    Extracts specified number of characters from the end of a string. If the string is shorter than
    requested number of characters, all characters of the string are returned.

    @return String containing extracted characters.
**/
string_t
right(
    string_t const & str,   ///< String to extract characters from.
    size_t           len    ///< Number of characters to extract.
) {
    return str.substr( str.size() >= len ? str.size() - len : 0, len );
}; // right

TEST(
    assert( right( "", 4 ) == "" );

    assert( right( "abc", 0 ) == "" );
    assert( right( "abc", 1 ) == "c" );
    assert( right( "abc", 2 ) == "bc" );
    assert( right( "abc", 3 ) == "abc" );
    assert( right( "abc", 4 ) == "abc" );

    if ( ENABLE_GNOME ) {
        assert( right( "эюя", 0 ) == "" );
        assert( right( "эюя", 1 ) == "я" );
        assert( right( "эюя", 2 ) == "юя" );
        assert( right( "эюя", 3 ) == "эюя" );
        assert( right( "эюя", 4 ) == "эюя" );
    }; // ENABLE_GNOME
);

/**
    Returns `true` if a string looks like an integer number. More precisely, returns `true` if the
    string comprises only digits (no leading or trailing spaces are allowed, no sign is allowed
    too), and `false` otherwise.
**/
bool
like_integer(
    string_t const & str
) {
    if ( str.empty() ) {
        return false;
    };
    size_t pos = ( str[ 0 ] == '+' or str[ 0 ] == '-' ) ? 1 : 0;
    if ( pos >= str.size() ) {
        return false;
    };
    return str.find_first_not_of( "0123456789", pos ) == npos;
}; // like_integer

TEST(
    assert( like_integer( "1" ) );
    assert( like_integer( "+12" ) );
    assert( like_integer( "-34" ) );
    assert( ! like_integer( "" ) );
    assert( ! like_integer( "a" ) );
    assert( ! like_integer( "я" ) );
    assert( ! like_integer( "1.1" ) );
    assert( ! like_integer( "+1.2" ) );
    assert( ! like_integer( "-3.4" ) );
    assert( ! like_integer( "1a" ) );
    assert( ! like_integer( "1я" ) );
    assert( ! like_integer( "a1" ) );
    assert( ! like_integer( "я1" ) );
    assert( ! like_integer( "1 " ) );
    assert( ! like_integer( " 1" ) );
);

/**
    Returns an uppercased version of string. If Gnome support is disabled, the function works only
    with ASCII characters.
**/
string_t
uc(
    string_t const & str
) {
    #if ENABLE_GNOME
        return str.uppercase();
    #else
        string_t res;
        res.reserve( str.size() );
        for ( size_t i = 0, end = str.size(); i < end; ++ i ) {
            res += toupper( str[ i ] );
        }; // for
        return res;
    #endif // ENABLE_GNOME
}; // uc

TEST(
    assert( uc( "" ) == "" );
    assert( uc( "1234567890-+=" ) == "1234567890-+=" );

    assert( uc( "abc" ) == "ABC" );
    assert( uc( "ABC" ) == "ABC" );
    if ( ENABLE_GNOME ) {
        assert( uc( "эюя" ) == "ЭЮЯ" );
        assert( uc( "ЭЮЯ" ) == "ЭЮЯ" );
    }; // ENABLE_GNOME
);

/**
    Returns an uppercased version of string. If Gnome support is disabled, the function works only
    with ASCII characters.
**/
string_t
lc(
    string_t const & str
) {
    #if ENABLE_GNOME
        return str.lowercase();
    #else
        string_t res;
        res.reserve( str.size() );
        for ( size_t i = 0, end = str.size(); i < end; ++ i ) {
            res += tolower( str[ i ] );
        }; // for
        return res;
    #endif // ENABLE_GNOME
}; // lc

TEST(
    assert( lc( "" ) == "" );
    assert( lc( "1234567890-+=" ) == "1234567890-+=" );

    assert( lc( "abc" ) == "abc" );
    assert( lc( "ABC" ) == "abc" );
    if ( ENABLE_GNOME ) {
        assert( lc( "эюя" ) == "эюя" );
        assert( lc( "ЭЮЯ" ) == "эюя" );
    }; // ENABLE_GNOME
);

string_t
chop(
    string_t const & str,   ///< String to chop suffix from.
    string_t const & sfx    ///< Suffix to chop.
) {
    if ( right( str, sfx.size() ) == sfx ) {
        return left( str, str.size() - sfx.size() );
    } else {
        return str;
    }; // if
}; // chop

TEST(
    assert( chop( "", "" ) == "" );

    assert( chop( "abc", "" ) == "abc" );
    assert( chop( "", "z" ) == "" );
    assert( chop( "abc", "z" ) == "abc" );
    assert( chop( "abcz", "z" ) == "abc" );
    assert( chop( "abczz", "z" ) == "abcz" ); // chops only once.

    assert( chop( "эюя", "" ) == "эюя" );
    assert( chop( "", "ё" ) == "" );
    assert( chop( "эюя", "ё" ) == "эюя" );
    assert( chop( "эюяё", "ё" ) == "эюя" );
    assert( chop( "эюяёё", "ё" ) == "эюяё" ); // chops only once.
);

string_t
q(
    string_t const & str
) {
    // TODO: Escape inner ` and ', and non-printable characters?
    return "`" + str + "'";
};

}; // namespace tapper

// end of file //
