/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/types.hpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Basic types defined here, like `time_t`, `key_t`, `key_range_t`.
**/

#ifndef _TYPES_HPP_
#define _TYPES_HPP_

#include <map>              // std::map
#include <memory>           // std::unique_ptr
#include <ostream>          // std::ostream

#include <cstddef>          // size_t
#include <cstdint>          // uint32_t

#include "reverse.hpp"
#include "string.hpp"

namespace tapper {

using std::size_t;

typedef std::uint32_t   uint_t;
template< typename obj_t > using ptr_t = std::unique_ptr< obj_t >;

/**
    Integer type to represent event's time. Could be either 32- or 64-bit integer.
**/
typedef uint_t  time_t;     // TODO: uint64_t?

/**
    Key code. Represented by `unt_t`, but actual range of key codes is expected to be much
    narrower, something like [1..255]. See `key_max` in `interceptor_t`.
**/
typedef uint_t key_t;

/**
    Key code range.
**/
struct key_range_t {
    key_range_t( uint_t _min = 1, uint_t _max = 255 );

    bool includes( uint_t code ) const;

    key_range_t & operator +=( key_t key );
    key_range_t & operator +=( key_range_t const & that );

    key_t  min;     ///< Lower bounary of range.
    key_t  max;     ///< Upper boundary of range.
}; // struct key_range_t

typedef ptr_t< key_range_t > key_range_p;

typedef std::map< string_t, key_t >  key_names_t;
typedef ptr_t< key_names_t >         key_names_p;
typedef reverse_map_t< key_names_t > key_codes_t;
typedef ptr_t< key_codes_t >         key_codes_p;

std::ostream &
operator <<(
    std::ostream & stream,
    key_range_t    range
);

}; // namespace tapper

#endif // _TYPES_HPP_

// end of file //
