/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/x.cpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    X Window System wrappers implementation.
**/

#include "x.hpp"

#include <cassert>

#include <unistd.h>
#include <linux/input-event-codes.h>    // BTN_LEFT, etc.

namespace tapper {
namespace x {

/**
    X key code offset. X key codes are biased libinput key codes. Offset is not documented but
    hardcoded in `xorg-x11-drv-libinput sources`, see `XORG_KEYCODE_OFFSET` macro in
    `xf86libinput.c`.
**/
static uint_t const key_code_offset = 8;

static XErrorHandler _standard_error_handler = nullptr;

// Some of `XErrorEvent` fiedls are of `unsigned char` type, and printed as characters. It is
// undesired behavior, let us fix it and pring them as integer numbers.
static
std::ostream &
operator <<(
    std::ostream &  stream,
    unsigned char   ch
) {
    stream << uint_t( ch );
    return stream;
}; // operator <<

static
int
error_handler(
    Display *       display,
    XErrorEvent *   event
) {
    #if 1
        char error_text[ 1000 ];
        XGetErrorText( event->display, event->error_code, error_text, sizeof( error_text ) );
        // TODO: How to decode opcode?
        // If I could decode opcode, I would not call standard error handler.
        ERR(
            "Display " << display << " reports error:\n"
            "    Display connection : " << XConnectionNumber( event->display ) << "\n"
            "    Request number     : " << event->serial << "\n"
            "    Opcode             : " << event->request_code << "." << event->minor_code << "\n"
            "    Error              : " << event->error_code << " (" << error_text << ")\n"
        );
    #endif
    #if 1
        if ( _standard_error_handler ) {
            _standard_error_handler( display, event );
        }; // if
    #endif
    return 0;
}; // error_handler

void
init(
) {
    Status ok = XInitThreads();
    if ( not ok ) {
        ERR( "Initializing X failed" );
    }; // if
    _standard_error_handler = XSetErrorHandler( error_handler );
    DBG( "error handler installed" );
}; // init

/** Converts X key code to libinput key code. **/
key_t
key(
    uint_t code
) {
    if ( code < key_code_offset ) {
        WRN( "Unexpected X key code " << code << "." );
        return 0;
    };
    return code - key_code_offset;
};

/** Converts X button code to libinput key code. **/
key_t
button(
    uint_t code
) {
    // cppcheck-suppress shadowFunction
    key_t key = 0;
    /*
        Function `btn_xorg2linux` from `x86libinput.c` implements the following scheme:

            Button1 → BTN_LEFT
            Button2 → BTN_MIDDLE
            Button3 → BTN_RIGHT
            ButtonN → ButtonN - 8 + BTN_SIDE

        It looks ok for the first 3 buttons, but for others it produces strange results: Button4 is
        mapped to 271 (unassigned code in BTN_MISC area), Button5 is mapped to 272 (BTN_LEFT, the
        same as Button1), etc.

        Let's follow this scheme for the first 3 buttons, and map Button4, Button5 and others to
        the keys following the first 3 buttons BTN_LEFT, BTN_MIDDLE, and BTN_RIGHT: BTN_SIDE,
        BTN_EXTRA, etc. It may be not the best solution, but it avoids flooding the log with
        warnings "unexpected X button" when user scrolls mouse wheel.
    */
    switch ( code ) {
        case Button1 : {
            key = BTN_LEFT;
        } break;
        case Button2 : {
            key = BTN_MIDDLE;
        } break;
        case Button3 : {
            key = BTN_RIGHT;
        } break;
        default : {
            if ( code <= button_max ) {
                key = BTN_MOUSE + code - Button1;
            } else {
                WRN( "Unexpected X button code " << code << "." );
            };
        };
    };
    return key;
};

// -------------------------------------------------------------------------------------------------
// display_t
// -------------------------------------------------------------------------------------------------

display_t::display_t(
) :
    OBJECT_T()
{
    TRACE();
    _rep = XOpenDisplay( nullptr );
    if ( not _rep ) {
        ERR( "Opening display " << q( XDisplayName( nullptr ) ) << " failed." );
    }; // if
    DBG(
        "opened, "
            << "name " << q( XDisplayString( _rep ) ) << ", "
            << "connection " << connection()
    );
}; // ctor

display_t::~display_t(
) {
    TRACE();
    CATCH_ALL(
        if ( _rep ) {
            Status ok = XCloseDisplay( _rep );
            // XCloseDisplay always returns 0. WTF??
            ok = ok || 1;       // Simple assignment causes cppcheck error.
            if ( not ok ) {
                ERR( "Closing display failed." );
            }; // if
            _rep = nullptr;
            DBG( "closed" );
        }; // if
    );
}; // dtor

display_t::operator Display *(
) const {
    assert( _rep );
    return _rep;
};

int
display_t::connection(
) const {
    return XConnectionNumber( _rep );
}; // connection

void
display_t::flush(
) {
    Status ok = XFlush( _rep );
    if ( not ok ) {
        ERR( "Flushing display failed." );
    }; // if
}; // flush

void
display_t::sync(
    bool discard
) {
    Status ok = XSync( _rep, discard );
    if ( not ok ) {
        ERR( "Synchronizing display failed." );
    }; // if
}; // sync

// -------------------------------------------------------------------------------------------------
// atom_t
// -------------------------------------------------------------------------------------------------

atom_t::atom_t(
    display_t const & display,
    Atom              atom
) :
    OBJECT_T(),
    _display( display ),
    _rep( atom )
{
}; // ctor

string_t
atom_t::name(
) {
    char * name = XGetAtomName( _display, _rep );
    // cppcheck-suppress nullPointerRedundantCheck
    if ( not name ) {
        ERR( "Getting atom name failed." );
    }; // if
    string_t res = name;
    XFree( name );
    return res;
}; // name

// =================================================================================================
// record
// =================================================================================================

/*
    Primary documentation:
    http://www.x.org/releases/X11R7.7/doc/libXtst/recordlib.html
*/

namespace record {

// -------------------------------------------------------------------------------------------------
// intercept_data_t
// -------------------------------------------------------------------------------------------------

intercept_data_t::intercept_data_t(
    XRecordInterceptData * _data
) :
    OBJECT_T(),
    data( _data )
{
}; // ctor

intercept_data_t::~intercept_data_t(
) {
    if ( data ) {
        XRecordFreeData( data ); // void function
        data = nullptr;
    }; // if
}; // dtor

// -------------------------------------------------------------------------------------------------
// range_t
// -------------------------------------------------------------------------------------------------

range_t::range_t(
    unsigned char first,
    unsigned char last
) :
    OBJECT_T()
{
    _rep = XRecordAllocRange();
    if ( not _rep ) {
        ERR( "Allocating record range failed." );
    }; // if
    _rep->device_events.first = first;
    _rep->device_events.last  = last;
    DBG( "allocated" );
}; // ctor

range_t::~range_t() {
    CATCH_ALL(
        if ( _rep ) {
            Status ok = XFree( _rep );
            if ( not ok ) {
                ERR( "Freeing record range failed." );
            }; // if
            DBG( "freed" );
            _rep = nullptr;
        }; // if
    );
}; // dtor

// -------------------------------------------------------------------------------------------------
// context_t
// -------------------------------------------------------------------------------------------------

context_t::context_t(
    on_intercept_t      on_intercept,
    int                 datum_flags,
    XRecordClientSpec   client
) :
    OBJECT_T(),
    _on_intercept( on_intercept ),
    _thread( _data )
{

    TRACE();

    // TODO: Move version check into a separate function?
    // TODO: Do we need to check version?
    int major = 0;
    int minor = 0;
    Status ok = XRecordQueryVersion( _ctrl, & major, & minor );
    if ( not ok ) {
        ERR( "Querying X record extension version failed." );
    }; // if
    DBG( "X record extension version " << major << "." << minor << "." );

    range_t         keys(       KeyPress,    KeyRelease );
    range_t         buttons( ButtonPress, ButtonRelease );
    XRecordRange *  ranges[ 2 ] = { keys._rep, buttons._rep };
    _handle = XRecordCreateContext(
        _ctrl,
            /*
                This request is typically executed by the recording client over its control
                connection to the X server.
                http://www.x.org/releases/X11R7.7/doc/libXtst/recordlib.html#XRecordCreateContext
            */
        datum_flags,
        & client,
        1,          // Number of clients.
        ranges,
        sizeof( ranges ) / sizeof( XRecordRange * )     // Number of ranges.
    );
    if ( _handle == None ) {
        ERR( "Creating record context failed." );
    }; // if
    DBG( "created" );

}; // ctor

context_t::~context_t(
) {
    TRACE();
    if ( _handle != None ) {
        CATCH_ALL(
            if ( _state == enabled ) {
                disable();
            }; // if
        );
        CATCH_ALL(
            Status ok = XRecordFreeContext( _ctrl, _handle );
            if ( not ok ) {
                ERR( "Freeing record context failed." );
            }; // if
            _handle = None;
            DBG( "freed" );
        );
    };
}; // dtor

display_t &
context_t::display(
) {
    return _ctrl;
};

void
context_t::enable(
) {
    TRACE();
    _ctrl.sync();
    /*
        Sync is a must. Otherwise XRecordEnableContextAsync will likely fail with error
        XRecordBadContext.
    */
    Status ok = XRecordEnableContextAsync(
        _data,
        _handle,
        interceptor,
        reinterpret_cast< XPointer >( this )
    );
    if ( not ok ) {
        ERR( "Enabling record context failed." );
    }; // if
    /*
        My experiments show interceptor is called once with event XRecordStartOfData from within
        XRecordEnableContextAsync, so context should be enabled now.
    */
    if ( _state != enabled ) {
        ERR( "Record context is not really enabled." );
    }; // if
    _thread.start();
}; // enable

void
context_t::disable(
) {
    TRACE();
    switch ( _state ) {
        case inited : {
            DBG( "not enabled, nothing do do" );
        } break;
        case enabled : {
            /*
                To direct the X server to halt the reporting of recorded protocol, the
                program executes XRecordDisableContext typically over its **control**
                connection to the X server.

                The XRecordDisableContext function disables context, stopping all
                recording over its data connection. Any complete protocol elements for
                context that were buffered in the server will be sent to the recording
                client rather than being discarded. If a program attempts to disable an
                XRecordContext that has not been enabled, no action will take place.

                http://www.x.org/releases/X11R7.7/doc/libXtst/recordlib.html#Disable_Context
            */
            Status ok = XRecordDisableContext( _ctrl, _handle );
            if ( ok ) {
                DBG( "disabled" );
            } else {
                ERR( "Disabling record context failed." );
            }; // if
            _ctrl.sync();   // Sync is a must.
            _thread.join();
            if ( _state != disabled ) {
                ERR( "Record context is not really disabled." );
            }; // if
        } break;
        case disabled : {
            DBG( "already disabled, nothing do do" );
        } break;
        default : {
            ERR( "Oops" );
        } break;
    }; // switch
}; // disable

void
context_t::interceptor(
    XPointer                closure,
    XRecordInterceptData *  data
) {
    /*
        (This code is executed by (non-initial) thread.)
        Xlib is a C library, it does not work with exception. If an exception leaked from this code,
        Xlib leaves an internal mutex locked, and program will hang. So we have to catch all the
        exceptions here.
    */
    context_t * context = reinterpret_cast< context_t * >( closure );
    // cppcheck-suppress shadowFunction
    THIS( context );
    TRACE();
    try {
        intercept_data_t intercept_data( data );
        switch ( data->category ) {
            case XRecordStartOfData : {
                // XServer sends such record at the very beginning.
                context->_state = enabled;
            } break;
            case XRecordEndOfData : {
                // XServer sends such record at the very end.
                context->_state = disabled;
            } break;
            case XRecordFromServer : {
                context->_on_intercept( data );
            } break;
        }; // switch
    } catch ( std::exception const & ex ) {
        // No need to warn: We just save the message now, an exception will be thrown when initial
        // thread calls `join`.
        context->_thread._what = ex.what();
        CATCH_ALL( posix::signal::kill( getpid(), SIGTERM ); );
    } catch ( ... ) {
        context->_thread._what = "Unknown exception occurred";
        CATCH_ALL( posix::signal::kill( getpid(), SIGTERM ); );
    }; // try
}; // interceptor

context_t::thread_t::thread_t(
    display_t & display
) :
    _display( display )
{
}; // ctor

void
context_t::thread_t::body(
) {
    TRACE();
    _display.sync();
}; // body

}; // namespace record

// -------------------------------------------------------------------------------------------------
// kb
// -------------------------------------------------------------------------------------------------

/*
    http://www.x.org/releases/X11R7.7/doc/libX11/XKB/xkblib.html
*/

kb_t::desc_t::desc_t(
    kb_t &  kb,
    uint_t  which
) :
    OBJECT_T(),
    _kb( kb ),
    _rep( XkbGetKeyboard( _kb.display(), which, XkbUseCoreKbd ) )
{
    if ( not _rep ) {
        ERR( "Getting keyboard description failed." );
    }; // if
}; // ctor

kb_t::desc_t::~desc_t(
) {
    if ( _rep ) {
        XkbFreeKeyboard( _rep, 0, True );
    }; // if
}; // dtor

key_range_t
kb_t::desc_t::key_range(
) const {
    return key_range_t( _rep->min_key_code, _rep->max_key_code );
}; // key_range_t

key_names_t
kb_t::desc_t::key_names(
) const {
    key_names_t names;
    for ( int code = _rep->min_key_code, end = _rep->max_key_code; code <= end; ++ code ) {
        // Key name is a fixed-length string padded with zeros.
        // We need to trim trailing zeros.
        char const * n = _rep->names->keys[ code ].name;
        size_t l = XkbKeyNameLength;
        while ( l > 0 and n[ l - 1 ] == 0 ) {
            -- l;
        }; // while
        auto const name = string_t( n, l );
        if ( not name.empty() ) {   // Ignore empty names.
            auto inserted = names.insert( { name, code } ).second;
            if ( not inserted ) {
                WRN( "Xkb reported non-unique key name " << q( name ) << "." );
            };
        };
    }; // for code
    return names;
}; // key_names

strings_t
kb_t::desc_t::group_names(
) const {
    strings_t group_names( layout_max );
    for ( size_t i = 0, end = layout_max; i < end; ++ i ) {
        if ( _rep->names->groups[ i ] != None ) {
            group_names[ i ] = atom_t( _kb.display(), _rep->names->groups[ i ] ).name();
        }; // if
    }; // for i
    return group_names;
}; // group_names

kb_t::kb_t(
    display_t & display
) :
    OBJECT_T(),
    _display( display )
{
    {
        int major = XkbMajorVersion;
        int minor = XkbMinorVersion;
        Bool ok = XkbLibraryVersion( & major, & minor );
        if ( not ok ) {
            ERR(
                "X keyboard extension: "
                    << "Incompatible client library version " << major << "." << minor << "; "
                    << "required version " << XkbMajorVersion << "." << XkbMinorVersion << "."
            );
        }; // if
        DBG( "X keyboard extension: client library version " << major << "." << minor << "." );
    }
    {
        int major = XkbMajorVersion;
        int minor = XkbMinorVersion;
        int dummy1, dummy2, dummy3;
        Bool ok = XkbQueryExtension( _display, & dummy1, & dummy2, & dummy3, & major, & minor );
        if ( not ok ) {
            ERR(
                "X keyboard extension: "
                "Incompatible server extension " << major << "." << minor << "; "
                    << "required version " << XkbMajorVersion << "." << XkbMinorVersion << "."
            );
        }; // if
        DBG( "X keyboard extension: server extension version " << major << "." << minor << "." );
    }
}; // ctor

display_t &
kb_t::display(
) {
    return _display;
}; // display

kb_t::autorepeat_rate_t
kb_t::autorepeat_rate(
) {
    autorepeat_rate_t rate;
    Bool ok = XkbGetAutoRepeatRate( _display, XkbUseCoreKbd, & rate.timeout, & rate.interval );
    if ( not ok ) {
        ERR( "Getting autorepeat rate failed." );
    }; // if
    DBG( "autorepeat rate: timeout = " << rate.timeout << ", interval = " << rate.interval );
    return rate;
}; // autorepeat_rate

void
kb_t::lock_group(
    uint_t  group
) {
    TRACE();
    Bool ok = XkbLockGroup( _display, XkbUseCoreKbd, group );
    if ( not ok ) {
        ERR( "Locking group " << ( group + 1 ) << " failed." );
    }; // if
    DBG( "group " << ( group + 1 ) << " locked" );
}; // lock_group

void
kb_t::bell(
) {
    Bool ok = XkbBell( _display, None, 100, None );
    if ( not ok ) {
        WRN( "Bell failed." );
    }; // if
}; // bell

kb_t::desc_p
kb_t::desc(
    uint_t  which
) {
    return desc_p( new desc_t( * this, which ) );
}; // desc

}; // namespace x
}; // namespace tapper

// end of file //
