/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/x.hpp

    Copyright © 2014, 2016—2017, 2019, 2020 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    X Window System wrappers interface.
**/

#ifndef _X_HPP_
#define _X_HPP_

#include <functional>

#include "base.hpp"

#include "posix.hpp"

#include <stdexcept>        // std::runtime_error

#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <X11/extensions/record.h>
#include <X11/extensions/XKB.h>

namespace tapper {
/// X Window System wrappers.
namespace x {

    static uint_t const layout_max = XkbNumKbdGroups;

    /** Minimal button code. **/
    static uint_t const button_min = Button1;

    /**
        Maximal button code. "X.h" header defines constants for 5 mouse buttons: Button1, ...,
        Button5. However, X can actually report more buttons (up to 255, because they use a byte
        for button code, and 0 is not used). However, "input-event-keycodes.h" reserves room for
        only 16 mouse buttons (from  BTN_MOUSE upto but not including BTN_JOYSTICK).

        TODO: X mouse button can be get through XGetPointerMapping.
    **/
    static uint_t const button_max = Button1 + 15;

    void init();

    key_t key( uint_t code );
    key_t button( uint_t code );

    // ---------------------------------------------------------------------------------------------
    // error_t
    // ---------------------------------------------------------------------------------------------

    /** X Window System exception class. **/
    class error_t : public std::runtime_error {
        public :
            using std::runtime_error::runtime_error;
    }; // class error_t

    // ---------------------------------------------------------------------------------------------
    // display_t
    // ---------------------------------------------------------------------------------------------

    /**
        `Display *` wrapper. Calls `XOpenDisplay` in constructor, and makes sure `XCloseDisplay` is
        called in destructor.
    **/
    class display_t : public object_t {

        public :

            explicit display_t();
            virtual ~display_t();

            operator Display *() const;

            int connection() const;

            void flush();
            void sync( bool discard = false );

        private :

            Display * _rep = nullptr;

    }; // class display_t

    typedef ptr_t< display_t > display_p;

    // ---------------------------------------------------------------------------------------------
    // atom_t
    // ---------------------------------------------------------------------------------------------

    class atom_t : public object_t {

        public :

            explicit atom_t( display_t const & display, Atom atom );

            string_t name();

        private:

            display_t const &   _display;
            Atom                _rep;

    }; // atom_t

    namespace record {

        // -----------------------------------------------------------------------------------------
        // intercept_data_t
        // -----------------------------------------------------------------------------------------

        /**
            `XRecordInterceptData *` wrapper. Makes sure `XRecordFreeData` is called.
        **/
        class intercept_data_t : public object_t {

            public :

                explicit intercept_data_t( XRecordInterceptData * data = nullptr );
                virtual ~intercept_data_t();

            public :

                XRecordInterceptData * data;

        }; // class intercept_data_t

        // -----------------------------------------------------------------------------------------
        // range_t
        // -----------------------------------------------------------------------------------------

        class range_t : public object_t {

            public :

                explicit range_t( unsigned char first, unsigned char last );
                virtual ~range_t();

            public :

                XRecordRange * _rep = nullptr;

        }; // class range_t

        // -----------------------------------------------------------------------------------------
        // context_t
        // -----------------------------------------------------------------------------------------

        class context_t : public object_t {

            public :

                typedef std::function< void( XRecordInterceptData const * ) > on_intercept_t;

                explicit context_t(
                    on_intercept_t    on_intercept,
                    int               datum_flags = 0,
                    XRecordClientSpec client      = XRecordAllClients
                );
                virtual ~context_t();

                display_t & display();

                void enable();
                void disable();

            private :

                static void interceptor( XPointer closure, XRecordInterceptData * data );

                class thread_t : public posix::thread_t {
                    friend class context_t;
                    public :
                        explicit thread_t( display_t & display );
                    protected :
                        virtual void body() override;
                    private :
                        display_t & _display;
                }; // thread_t

            private :

                enum state_t {
                    inited,
                    enabled,
                    disabled
                }; // enum state_t

                on_intercept_t  _on_intercept;
                state_t         _state  { inited };
                XRecordContext  _handle { None };
                display_t       _ctrl;
                display_t       _data;
                thread_t        _thread;

        }; // class context_t

    }; // namespace record

    // ---------------------------------------------------------------------------------------------
    // kb_t
    // ---------------------------------------------------------------------------------------------

    class kb_t : public object_t {

        public :

            struct autorepeat_rate_t {
                uint_t  timeout  = 0;
                uint_t  interval = 0;
            }; // struct autorepeat_rate_t

            /** Keyboard description. **/
            class desc_t : public object_t {

                friend class kb_t;

                private :

                    explicit desc_t( kb_t & kb, uint_t which );

                public :

                    virtual ~desc_t();

                    key_range_t     key_range()     const;
                    key_names_t     key_names()     const;
                    strings_t       group_names()   const;

                private :

                    kb_t &      _kb;
                    XkbDescPtr  _rep = nullptr;

            }; // class desc_t

            typedef ptr_t< desc_t > desc_p;

        public :

            explicit kb_t( display_t & display );

            display_t &       display();
            autorepeat_rate_t autorepeat_rate();
            void              lock_group( uint_t group );
            void              bell();
            desc_p            desc( uint_t which = XkbAllComponentsMask );

        private :

            display_t &     _display;

    }; // class kb_t

    typedef ptr_t< kb_t >   kb_p;

}; // namespace x
}; // namespace tapper

#endif // _X_HPP_

// end of file //
